#ifdef WIN32

#define _CRT_SECURE_NO_WARNINGS

#include <Windows.h>
#include <io.h>
#include <direct.h>

#define open _open
#define read _read
#define close _close
#define alloca _alloca
#define getcwd _getcwd
#define lseek _lseek
#define chdir _chdir
#define access _access
#define unlink _unlink

//#define WIFEXITED(status) 1
//#define WEXITSTATUS(status) status
#define F_OK 1

#endif


#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifndef WIN32
#include <sys/wait.h>
#endif
#include <fcntl.h>
#include <time.h>
#include <stdio.h>
#include <string.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include <stdlib.h>
#include <ctype.h>

#include "xmalloc.h"
#include "util.h"
#include "object.h"
#include "code.h"
#include "source.h"
#include "detect.h"
#include "ta.h"
#include "data.h"

// #define SANITY

int			g_hypertext_cnt      = 0;
char		*g_html_dirP         = 0;
char		*g_ta_fileP          = 0;
int			g_match_lines        = 1;
int			g_min_match_sequence = 15;
int			g_min_function_sequence = 10;
int			g_ignore_macros      = 0;
int			g_mismatch_cost      = -1;
int			g_same_cost          = 1;
int			g_comparisons        = 0;
int			g_report_c           = 1;
int			g_report_assembler   = 1;
int			g_debug_dwarf        = 0;
int			g_debug_code         = 0;
int			g_use_frame_address  = 0;
int			g_use_temp_address   = 0;
int			g_top_clones         = 1024;
int			g_top_cols           = 8;
int			g_hillclimb			 = 0;
int			g_iterations         = 0x7FFFFFFF;
int			g_correlate          = 0;
int			g_best_weight        = 0;

int			g_read               = 0;
int			g_compiled_ok        = 0;
int			g_compiled_bad       = 0;

long long	g_original_matches   = 0;
long long	g_extra_matches      = 0;
int			g_max_blocks_seen    = 0;
int			g_max_iterations_seen= -1;

char		**g_envPP;

FILE		*stdinF              = stdin;
FILE		*stdoutF             = stdout;
FILE		*redirectF           = stdoutF;
FILE		*taF                 = 0;
FILE		*tempF				 = 0;

int			g_verbose        = 0;
int			g_unlink         = 1;
int			g_skip_compile   = 0;
int			g_stdin_lineno   = 0;

#define SUFFIX_s 	0
#define SUFFIX_S    1
#define SUFFIX_c 	2
#define SUFFIX_C 	3
#define SUFFIX_cc 	4
#define SUFFIX_cxx	5
#define SUFFIX_cpp	6
#define SUFFIX_cpp1 7
#define SUFFIX_none 8

/* Must agree with SUFFIX_* */

static const char *suffixes[] =
{
	"s",
	"S",
	"c",
	"C",
	"cc",
	"cxx",
	"cpp",
	"c++",
	0
};


#ifdef WIN32

#include <process.h>

static int
call_cc(char *argv[])
{
	int		ret;
	int 	i;
	char	*P;

	if (g_verbose) {
		fprintf(stderr, "%4d: ", g_stdin_lineno);
		for (i = 0; (P = argv[i]); ++i) {
			fputs(P, stderr);
			fputc(' ', stderr);
		}
		fputc('\n', stderr);
	}
	
	ret = _spawnvpe(_P_WAIT, argv[0], argv+1, g_envPP);
	if (ret) {
		if (!g_verbose) {
			for (i = 0; P = argv[i]; ++i) {
				fputs(P, stderr);
				fputc(' ', stderr);
		}	}
		fprintf(stderr, "returned %d\n", ret);
	} 
	return(ret);
}
#else
static int
call_cc(char *argv[])
{
	int		ret, pid;
	int		status;
	int 	i;

	if (g_verbose) {
		fprintf(stderr, "%4d: ", g_stdin_lineno);
		for (i = 0; argv[i]; ++i) {
			fprintf(stderr, "%s ", argv[i]);
		}
		fprintf(stderr,"\n");
	}
	
	ret = fork();
	switch (ret) {
	case 0:
		execve(argv[0], argv, g_envPP);
		fprintf(stderr,"** Line %d: Return from execve %s\n", g_stdin_lineno, argv[0]);
		error_exit(1);
	case -1:
		fprintf(stderr,"** Line %d: Fork to cc failed\n", g_stdin_lineno);
		perror("");
		error_exit(2);
	}
	pid = wait(&status);
	if (pid != ret) {
		fprintf(stderr,"** Line %d: Wait for cc failed\n", g_stdin_lineno);
		perror("");
		error_exit(3);
	}
	if (!WIFEXITED(status)) {
		fprintf(stderr,"** Line %d: cc did not terminate normally\n", g_stdin_lineno);
		return(-1);
	}
	return(WEXITSTATUS(status));
}
#endif

static void
get_environment(int argc, char **argv)
{
	int		i, val;
	char	*P;

	for (i = 1; i < argc;) {
		P = argv[i++];
		if (*P == '-') {
			switch (P[1]) {
			case 'a':
				g_match_lines = 0;
				if (P[2]) {
					P += 2;
				} else {
					if (argc <= i) {
						break;
					}
					P = argv[i++];
				}
				switch (P[0]) {
				case 'l':
					g_match_lines = 1;	// Match start of source lines
					break;
				}
				break;
			case 'b':
				g_best_weight = 1;
				break;
			case 'd':
				g_debug_dwarf = 1;
				break;
			case 'D':
				g_debug_code  = 1;
				break;
			case 'f':
				g_use_frame_address = 1;
				break;
			case 'F':
				g_use_temp_address = 1;
				break;
			case 'h':
				if (P[2]) {
					g_html_dirP = P + 2;
				} else {
					if (argc <= i) {
						break;
					}
					g_html_dirP = argv[i++];
				}
#ifdef WIN32
				CreateDirectoryA(g_html_dirP, 0);
#else
				mkdir(g_html_dirP, 0755);
#endif
				sprintf(g_filename, "%s/index.html", g_html_dirP);
				redirectF = open_file(g_filename);
				stdoutF = redirectF;
				break;
			case 't':
				if (P[2]) {
					g_ta_fileP = P + 2;
				} else {
					if (argc <= i) {
						break;
					}
					g_ta_fileP = argv[i++];
				}
				break;
			case 'k':
				g_unlink = 0;
				break;
			case 'S':
				g_skip_compile = 1;
				break;
			case 'T':
				if (P[2]) {
					P += 2;
				} else {
					if (argc <= i) {
						break;
					}
					P = argv[i++];
				}
				val = atoi(P);
				g_top_clones = val;
				break;
			case 'C':
				if (P[2]) {
					P += 2;
				} else {
					if (argc <= i) {
						break;
					}
					P = argv[i++];
				}
				val = atoi(P);
				if (val <= 0) {
					fprintf(stderr, "-C %d not positive\n", val);
					error_exit(1);
				}
				g_top_cols = val;
				break;
			case 'l':
				if (P[2]) {
					P += 2;
				} else {
					if (argc <= i) {
						break;
					}
					P = argv[i++];
				}
				val = atoi(P);
				if (val < 1) {
					fprintf(stderr, "-l/ength must be > 0\n");
					error_exit(1);
				}
				g_min_match_sequence = val;
				break;
			case 'L':
				if (P[2]) {
					P += 2;
				} else {
					if (argc <= i) {
						break;
					}
					P = argv[i++];
				}
				val = atoi(P);
				if (val < 1) {
					fprintf(stderr, "-L/ength must be > 0\n");
					error_exit(1);
				}
				g_min_function_sequence = val;
				break;
			case 'M':
				g_ignore_macros = 1;
				break;
			case 'm':
				if (P[2]) {
					P += 2;
				} else {
					if (argc <= i) {
						break;
					}
					P = argv[i++];
				}
				val = atoi(P);
				if (val < 0) {
					fprintf(stderr, "-m/ismatch cost must be => 0\n");
					error_exit(1);
				}
				g_mismatch_cost = -val;
				break;
			case 's':
				if (P[2]) {
					P += 2;
				} else {
					if (argc <= i) {
						break;
					}
					P = argv[i++];
				}
				val = atoi(P);
				if (val < 0) {
					fprintf(stderr, "-s/ame cost must be => 0\n");
					error_exit(1);
				}
				g_same_cost = val;
				break;
			case 'v':
				g_verbose = 1;
				break;
			case 'r':
				g_report_c = 0;
				break;
			case 'R':
				g_report_assembler = 0;
				break;
			case 'u':
				if (P[2]) {
					P += 2;
				} else {
					if (argc <= i) {
						break;
					}
					P = argv[i++];
				}
				val = atoi(P);
				if (val < 0) {
					fprintf(stderr, "-Use option must be non-negative\n");
					error_exit(1);
				}
				g_hillclimb = val;
				break;
			case 'i':
				if (P[2]) {
					P += 2;
				} else {
					if (argc <= i) {
						break;
					}
					P = argv[i++];
				}
				val = atoi(P);
				if (val < 0) {
					fprintf(stderr, "-i/terate option must be non-negative\n");
					error_exit(1);
				}
				g_iterations = val;
				break;
			case 'z':
				g_correlate = 1;
				break;

			default:
				fprintf(stderr, "Unknown option \"%s\"\n\n", P);
				fputs("Options:\n\n"
					"  -a/ll [i/nstructions|l/ines]\n"
					"  -b/estweight\n"
					"  -C/ols <number>\n"
					"  -d/ebug\n"
					"  -D/ebug\n"
					"  -f/rame address\n"
					"  -F/rame temp address\n"
					"  -h/tml <directory>\n"
					"  -k/eep assembler\n"
					"  -l/ength_assember <min>\n"
					"  -M/acro detection\n"
					"  -m/ismatch <cost>\n"
					"  -r/emove source reporting\n"
					"  -R/emove assembler reporting\n"
					"  -s/ame <cost>\n"
					"  -S/kip compile\n"
					"  -t/a <file>\n"
					"  -T/op <number>\n"
					"  -u/se <number> [-i/terate <number>]\n"
					"  -v/erbose\n"
					, stderr);
				error_exit(1);
	}	}	}
	if (!g_html_dirP || (!g_report_assembler && !g_report_c)) {
		g_top_clones = 0;
	}
	return;
}

static void
print_title(int argc, char *argv[])
{
	int	i;

	fputs( HTML_HEADER "<title>", stdoutF);
	fprintf(stdoutF, "ACD Version " VERSION " Compiled " __DATE__ " "  __TIME__ );
	fputs("</title>\n<body>\n", stdoutF);
	fprintf(stdoutF, "<h3>Arguments to %s</h3>\n<p>\n", argv[0]);
	for (i = 1; i < argc; ++i) {
		fprintf(stdoutF,"\"%s\" ", argv[i]);
	}
	if (g_top_clones > 0) {
		fprintf(stdoutF, "<p>\n<a href=top.html>Top %d clones </a>\n", g_top_clones);
	}
	if (g_correlate > 0) {
		fprintf(stdoutF, "<p>\n<a href=correlated.html>Correlations</a>\n");
	}
}

static int
parse_command(char *commandP, char **argv1P)
{
	char	*startP, *endP, *nextP;
	int		i, c;

	startP = commandP;
	if (*startP != '"') {
		fprintf(stderr, "** Line %d: Missing cwd start '\"' >%s\n", g_stdin_lineno, startP);
		return(-1);
	}

	for (i = -1; startP; ++i) {
		for (endP = nextP = ++startP; (c = *nextP) != '"'; ++nextP) {
			switch (c) {
			case 0:
				fprintf(stderr, "** Line %d: Missing end '\" >%s'\n", g_stdin_lineno, startP);
				return(-1);
			case '\\':
				c = *(++nextP);
			default:
				*endP++ = c;
		}	}
		*endP = 0;
		if (i < 0) {
			cwdP = startP;
			if (chdir(cwdP)) {
				fprintf(stderr,"** Line %d: Can't change directory to %s\n", g_stdin_lineno, cwdP);
				perror("");
				return(-1);
			}
		} else {
			argv1P[i] = startP;
		}
		startP = strchr(nextP+1, '"');
	}
	if (i < 1) {
		fprintf(stderr, "** Line %d: Missing cmd '\" >%s'\n", g_stdin_lineno, commandP);
		return(-1);
	}
	argv1P[i]   = 0;
	return(i);
}

static int
process_command(int argc2, char **argv1P)
{
	static char	filename1[1024];
	static char	*compilePP[2048];

	struct stat	source_stat;
	struct stat assembler_stat;

	int			compile;

	char 	*programP, *compileP, *targetP, *assemblerP, *absolute_pathP;
	char	*P, *P1;
	const char	*suffixP;
	int		i, suffix, unlink_me, lth, ret, argc1;
	int		seen_dash_S, seen_dash_E, running_as, skip_compile;
	int		c2;
	Csource	*sourceP;

	/* Find the source and target files */

	seen_dash_S = 0;
	seen_dash_E = 0;
	targetP     = 0;
	assemblerP  = 0;
	compile     = 0;
	running_as  = 0;
	skip_compile= g_skip_compile;
	programP    = argv1P[0];
	lth         = strlen(programP);
	if (lth > 3) {
		if (!strcmp(programP+lth-3, "/as")) {
			running_as = 1;
		} else if (lth > 4 && !strcmp(programP+lth-4, "/gas")) {
			running_as = 1;
	}	}
	
	if (running_as) {
		for (argc1 = i = 1; (P = argv1P[argc1]); ++argc1) {
			if (argc1 >= argc2) {
				fprintf(stderr, "** lineno %d: Can't scan args for %s\n", g_stdin_lineno, programP);
				return(-1);
			}
			if (*P != '-') {	
				if (!strcmp(P, "/dev/null")) {
					return(0);
				}
				compilePP[compile++] = P;
				continue;		// Discard (to add back in later)
			}
			switch(P[1]) {
			case '-':
				if (!strcmp(P,"--defsym")) {
					++argc1;	// Discard two arguments
				}
				continue;
			case 'I':
			case 'o':
				if (!P[2]) {
					++argc1;
				}
			case 'a':
			case 'D':
			case 'f':
			case 'g':
			case 'J':
			case 'K':
			case 'L':
			case 'n':
			case 'R':
			case 'v':
			case 'W':
			case 'w':
			case 'x':
			case 'Z':
				continue;
			}
			fprintf(stderr, "** Line %d: Unknown %s option '%s'\n", g_stdin_lineno, programP, P);
			goto bad;
		}
	} else {
		for (argc1 = i = 1; (P = argv1P[argc1]); ++argc1) {
			if (argc1 >= argc2) {
				fprintf(stderr, "** lineno %d: Can't scan args for %s\n", g_stdin_lineno, programP);
				return(-1);
			}
			if (*P != '-') {	
				if (!strcmp(P, "/dev/null")) {
					return(0);
				}
				compilePP[compile++] = P;
				continue;		// Discard (to add back in later)
			}
			c2 = P[2];
			switch(P[1]) {
			case 'a':
				if (!strcmp(P, "-aux-info")) { 		// Discard 2 args
					++argc1;
					continue;
				}
			case 'B':
			case 'n':
			case 't':
			case 'Y':
				argv1P[i++] = P;
				continue;
			case 'f':
				// Need frame pointers for dwarf
				if (strcmp(P,"-fomit-frame-pointer")) {
					argv1P[i++] = P;
				}
				continue;
			case 'D':
			case 'I':
			case 'U':
				if (!c2) {
					argv1P[i++] = P;
					P           = argv1P[++argc1];
				}
				argv1P[i++] = P;
				continue;
			case 'b':
				if (!c2) {
					++argc1;						// Discard 2 args
					continue;
				}
				argv1P[i++] = P;
			case 'c':
			case 'C':
			case 'g':
			case 'h':
			case '-':
			case 'M':
			case 'O':
			case 'P':
			case 'Q':
			case 'r':
			case 'v':
			case 'w':
			case 'l':
			case 'L':
				continue;							// Discard
			case 'd':
				if (c2 == 'D' || c2 == 'I' || c2 == 'M' || c2 == 'N') {
					argv1P[i++] = P;
				}
				continue;
			case 'E':
				if (!c2) {
					seen_dash_E = 1;
					continue;
				}
				argv1P[i++] = P;
				continue;
			case 'G':
			case 'm':
				if (!c2) {
					++argc1;
				}
				continue;
			case 'i':
				if (!strcmp(P, "-idirafter")     ||
					!strcmp(P, "-imacros")       ||
					!strcmp(P, "-include")       ||
					!strcmp(P, "-iprefix")       ||
					!strcmp(P, "-iwithprefix")   ||
					!strcmp(P, "-isystem")) {
					argv1P[i++] = P;
					P           = argv1P[++argc1];
				}
				argv1P[i++] = P;
				continue;
			case 'o':	// <file>
				if (!c2) {
					targetP = argv1P[++argc1];	// Discard parameter
					continue;
				}
				targetP = P + 2;
				continue;
			case 'p':
				if (strcmp(P, "-print")) {		// Not -print
					argv1P[i++] = P;
				} else {
					if (!strcmp(P, "-print-file-name")) {
						return(0);
				}	}
				continue;
			case 'S':
				if (!c2) {
					seen_dash_S = 1;
					continue;	
				}
				break;
			case 's':
				if (!strcmp(P, "-std")) {
					argv1P[i++] = P;
				}
				continue;
			case 'u':
			case 'V':
				if (!c2) {
					argv1P[i++] = P;
					P           = argv1P[++argc1];
				}
				argv1P[i++] = P;
				continue;
			case 'W':
				if (!strcmp(P, "-Wp,")) {
					if (!strcmp(P, "-Wp,-M")) {
						argv1P[i++] = P;
				}	}
				continue;
			case 'x':
				if (c2) {
					argv1P[i++] = P;
					continue;
				}
			case 'X':
				argv1P[i++] = P;
				argv1P[i++] = argv1P[++argc1];
				continue;
			case '#':
				if (!strcmp(P, "-###")) {
					continue;
				}
				break;
			}
			/* If the compiler was not compiling we shouldn't */
			fprintf(stderr, "** Line %d: Unknown compiler option '%s'\n", g_stdin_lineno, P);
			goto bad;
		}
	}

	argc1 = i;

	compilePP[compile] = 0;

	if (seen_dash_S) {
		assemblerP = targetP;
	}

	for (compile = 0; (compileP = compilePP[compile]); ++compile) {
		P1 = strrchr(compileP, '.');
		if (!P1) {
			suffix = SUFFIX_none;
		} else {
			++P1;
			for (suffix= 0;; ++suffix) {
			
				if (!(suffixP = suffixes[suffix])) {
					goto next_command;
				}
				if (!strcmp(P1, suffixP)) {
					break;
		}	}	}
	
		unlink_me  = 0;
		if (suffix == SUFFIX_S && running_as) {
			suffix = SUFFIX_s;
		}
		switch (suffix) {
		case SUFFIX_cxx:
		case SUFFIX_cpp:
		case SUFFIX_cpp1:
		case SUFFIX_c:
		case SUFFIX_C:
		case SUFFIX_cc:
			if (seen_dash_E) {
				goto next_command;
			}
		case SUFFIX_S:
			if (!assemblerP) {
				lth = P1-compileP;
				memcpy(filename1, compileP, lth);
				strcpy(filename1+lth, "s");
				assemblerP = filename1;
			}
			argv1P[argc1++] = compileP;
			argv1P[argc1]   = 0;
			argv1P         -= 4;
			argv1P[0]       = argv1P[4];
			argv1P[1]       = "-o";
			argv1P[2]       = assemblerP;
			argv1P[3]       = "-g";
			if (suffix == SUFFIX_S) {
				argv1P[4]   = "-E";
			} else {
				argv1P[4]   = "-S";
			}
		
			if (g_unlink || skip_compile) {
				if (access(assemblerP, F_OK)) {
					// Looks like file does not exist
					unlink_me    = g_unlink;
					skip_compile = 0;
				} else if (skip_compile) {
					if (!access(compileP, F_OK)) {
						// Looks like the source exists
						if (stat(compileP, &source_stat) ||
                            stat(assemblerP, &assembler_stat) ||
							source_stat.st_mtime >= assembler_stat.st_mtime) {
							skip_compile = 0;
			}	}	}	}
	
			if (!skip_compile) {
				ret = call_cc(argv1P);
				if (ret) {
					++g_compiled_bad;
					fprintf(stderr, "** Line %d: Failed to compile %s\n", g_stdin_lineno, compileP);
					unlink_me = 0;
					goto next_command;
				}
				++g_compiled_ok;
			}
			goto read_assembler;
	
		case SUFFIX_s:
			assemblerP     = compileP;
read_assembler:
			absolute_pathP = absolute_path(compileP);
			sourceP        = new Csource(absolute_pathP);
			if (!sourceP) {
				outofmemory();
				goto bad;
			}
	
			absolute_pathP = absolute_path(assemblerP);
			if (sourceP->load(absolute_pathP)) {
				goto next_command;
			}
	
			if (!unlink_me) {
				break;
			}
	
			unlink_me = 0;
	
			/* Sanity checks follow */
			if (assemblerP != filename1) {
				break;
			} 
			lth = strlen(assemblerP);
			if (lth < 3 || assemblerP[lth-2] != '.' || assemblerP[lth-1] != 's') {
				fprintf(stderr, "Line %d: May not delete %s\n", g_stdin_lineno, assemblerP);
				error_exit(5);
				goto bad;
			}
	
			/* Discard this assembly file when done */
	
			if (unlink(assemblerP)) {
				fprintf(stderr, "Line %d: Can't delete %s\n", g_stdin_lineno, assemblerP);
				perror("");
				break;
			}
			break;
		}
next_command:
		continue;
	}
	return(0);
bad:
	return(-1);
}

// Leave room for -g -S -o filename
#define ExtraSlots 4

#ifdef WIN32
#define clock_t time_t
#endif

int
main(int argc, char *argv[], char *env[])
{
	static char	*argv1[2048];
	
	char		**argv1P;
	int 		i;
	char		*stdin_bufferP;
	int			stdin_buffer_lth;
	clock_t		start_clock, read_clock, end_clock;
	int			hypertext_cnt;
	long long	total_matches;
	double		percentage;
	
	start_clock = time(0);

	g_envPP = env;
	get_environment(argc, argv);
	initialise_detect();
	init_lookup_chars();

	print_title(argc, argv);
	ta_header();
	fprintf(stdoutF, "<h3>Processing</h3>");

	stdin_buffer_lth = 256;
	stdin_bufferP    = (char *) Xmalloc(stdin_buffer_lth);
	while (getline1(&stdin_bufferP, &stdin_buffer_lth, stdin) >= 0) {
		argv1P = argv1 + ExtraSlots;
		++g_stdin_lineno;
		i = parse_command(stdin_bufferP, argv1P);
		if (i > 0) {
			process_command(i, argv1P);
	}	}

	fprintf(stderr, "\n\nRead %d Compiled %d Failed %d\n", g_read, g_compiled_ok, g_compiled_bad);
	// Now perform clone detection

	read_clock = time(0);

	start_report();
	init_data();
	hypertext_cnt = g_hypertext_cnt;
	detect();
	finalize_report();

	end_clock = time(0);
	total_matches = g_original_matches + g_extra_matches;
	percentage    = ((double) g_extra_matches)/((double) total_matches);
	percentage   *= 100.0;

	fprintf(redirectF, "</table>\n<p>\n<table>\n<tr><td>Start</td><td>%s</td></tr>\n", 	ctime(&start_clock));
	fprintf(redirectF, "<tr><td>Read</td><td align=right>%s</td></tr>\n", 	ctime(&read_clock));
	fprintf(redirectF, "<tr><td>End</td><td align=right>%s</td></tr>\n", 	ctime(&end_clock));
	fprintf(redirectF, "<tr><td>Elapse</td><td align=right>%d</td></tr>\n", end_clock-read_clock);
	fprintf(redirectF, "<tr><td>Sources</td><td align=right>%d</td></tr>\n", g_read);
	fprintf(redirectF, "<tr><td>Compiled</td><td align=right>%d</td></tr>\n", g_compiled_ok);
	if (g_compiled_bad) {
		fprintf(redirectF, "<tr><td>Uncompilable</td><td align=right>%d</td></tr>\n", g_compiled_bad);
	}
	fprintf(redirectF, "<tr><td>Comparisons</td><td align=right>%d</td></tr>\n", g_comparisons);
	fprintf(redirectF, "<tr><td>Clones detected</td><td align=right>%d</td></tr>\n", g_hypertext_cnt - hypertext_cnt);
	fprintf(redirectF, "<tr><td>Matches detected</td><td align=right>%lld</td></tr>\n", total_matches);
	if (g_hillclimb >= 0) {
		fprintf(redirectF, "<tr><td>Hill climbing</td><td align=right>%lld</td></tr>\n", g_extra_matches);
		fprintf(redirectF, "<tr><td>Max blocks</td><td align=right>%d</td></tr>\n", g_max_blocks_seen);
		fprintf(redirectF, "<tr><td>Max iterations</td><td align=right>%d</td></tr>\n", g_max_iterations_seen+1);
		fprintf(redirectF, "<tr><td>Percentage</td><td align=right>%f%%</td></tr>\n", percentage );
	}

	fprintf(redirectF, "</table>\n<p>\n<i>End of Report</i>\n</body>\n</html>\n");

	ta_trailer();
	fprintf(stderr, "Done secs=%d clones=%d matches=%lld",
				end_clock - read_clock,
				g_hypertext_cnt - hypertext_cnt,
				total_matches);
	if (g_hillclimb >= 0) {
		fprintf(stderr, " -u %d extra=%lld iterations=%d\n", 
				g_hillclimb,
				g_extra_matches,
				g_max_iterations_seen+1);
	}
	fprintf(stderr, "\n");
	return(0);
}
